package org.bitbucket.openisoj;

import org.bitbucket.openisoj.exceptions.FieldDescriptorException;
import org.bitbucket.openisoj.exceptions.FieldFormatException;
import org.bitbucket.openisoj.fieldvalidator.FieldValidators;
import org.bitbucket.openisoj.fieldvalidator.HexFieldValidator;
import org.bitbucket.openisoj.fieldvalidator.IFieldValidator;
import org.bitbucket.openisoj.fieldvalidator.NumericFieldValidator;
import org.bitbucket.openisoj.formatter.AsciiFormatter;
import org.bitbucket.openisoj.formatter.BcdFormatter;
import org.bitbucket.openisoj.formatter.BinaryFormatter;
import org.bitbucket.openisoj.formatter.Formatters;
import org.bitbucket.openisoj.formatter.IFormatter;
import org.bitbucket.openisoj.lengthformatters.FixedLengthFormatter;
import org.bitbucket.openisoj.lengthformatters.ILengthFormatter;
import org.bitbucket.openisoj.lengthformatters.VariableLengthFormatter;

public class FieldDescriptor implements IFieldDescriptor {

	public static IFieldDescriptor getAsciiAlphaNumeric(int length) throws Exception {
		return new FieldDescriptor(new FixedLengthFormatter(length), FieldValidators.getAnsp(),
				new PadRightSetAdjuster(length, ' '));
	}

	public static IFieldDescriptor getAsciiAns(int length) throws Exception {
		return new FieldDescriptor(new FixedLengthFormatter(length), FieldValidators.getAns(), null);
	}

	public static IFieldDescriptor getAsciiN(int length) throws Exception {
		return new FieldDescriptor(new FixedLengthFormatter(length), FieldValidators.getN(), null);
	}

	public static IFieldDescriptor getAsciiAmount(int length) throws Exception {
		return new FieldDescriptor(new FixedLengthFormatter(length), FieldValidators.getRev87Amount(),
				new PadLeftSetAdjuster(length, '0'));
	}

	public static IFieldDescriptor getAsciiFixed(int packedLength, IFieldValidator validator) throws Exception {
		return new FieldDescriptor(new FixedLengthFormatter(packedLength), validator, Formatters.getAscii(), null);
	}

	public static IFieldDescriptor getAsciiLLCharacter(int maxLength) throws Exception {
		return getAsciiVar(2, maxLength, FieldValidators.getAns());
	}

	public static IFieldDescriptor getAsciiLLLBinary(int packedLength) throws Exception {
		return new FieldDescriptor(new VariableLengthFormatter(3, packedLength), FieldValidators.getHex(),
				Formatters.getBinary(), null);
	}

	public static IFieldDescriptor getAsciiLLLCharacter(int maxLength) throws Exception {
		return getAsciiVar(3, maxLength, FieldValidators.getAns());
	}

	public static IFieldDescriptor getAsciiLLLNumeric(int maxLength) throws Exception {
		return getAsciiVar(3, maxLength, FieldValidators.getN());
	}

	public static IFieldDescriptor getAsciiLLNumeric(int maxLength) throws Exception {
		return getAsciiVar(2, maxLength, FieldValidators.getN());
	}

	public static IFieldDescriptor getAsciiNumeric(int length) throws Exception {
		return new FieldDescriptor(new FixedLengthFormatter(length), FieldValidators.getN(), new PadLeftSetAdjuster(
				length, '0'));
	}

	public static IFieldDescriptor getAsciiVar(int lengthIndicator, int maxLength, IFieldValidator validator)
			throws Exception {
		return new FieldDescriptor(new VariableLengthFormatter(lengthIndicator, maxLength), validator, null);
	}

	public static IFieldDescriptor getBcdFixed(int packedLength) throws Exception {
		return new FieldDescriptor(new FixedLengthFormatter(packedLength), FieldValidators.getN(), Formatters.getBcd(),
				new PadLeftSetAdjuster(packedLength * 2, '0'));
	}

	public static IFieldDescriptor getBcdVar(int lengthIndicator, int maxLength) throws Exception {
		return new FieldDescriptor(new VariableLengthFormatter(lengthIndicator, maxLength, new BcdFormatter()),
				FieldValidators.getN(), Formatters.getBcd(), null);
	}

	public static IFieldDescriptor getBinaryFixed(int packedLength) throws Exception {
		return new FieldDescriptor(new FixedLengthFormatter(packedLength), FieldValidators.getHex(),
				Formatters.getBinary(), null);
	}

	public static IFieldDescriptor getPanMask(IFieldDescriptor decoratedFieldDescriptor) {
		return new PanMaskDecorator(decoratedFieldDescriptor);
	}

	protected ILengthFormatter lengthFormatter;
	protected IFieldValidator validator;
	protected IFormatter formatter;
	protected Adjuster adjuster;

	public FieldDescriptor(ILengthFormatter lengthFormatter, IFieldValidator validator) throws Exception {
		this(lengthFormatter, validator, new AsciiFormatter(), null);
	}

	public FieldDescriptor(ILengthFormatter lengthFormatter, IFieldValidator validator, Adjuster adjuster)
			throws Exception {
		this(lengthFormatter, validator, new AsciiFormatter(), adjuster);
	}

	public FieldDescriptor(ILengthFormatter lengthFormatter, IFieldValidator validator, IFormatter formatter,
			Adjuster adjuster) throws Exception {
		if (formatter instanceof BinaryFormatter && !(validator instanceof HexFieldValidator))
			throw new FieldDescriptorException("A Binary field must have a hex validator");

		if (formatter instanceof BcdFormatter && !(validator instanceof NumericFieldValidator))
			throw new FieldDescriptorException("A BCD field must have a numeric validator");

		this.lengthFormatter = lengthFormatter;
		this.validator = validator;
		this.formatter = formatter;
		this.adjuster = adjuster;

	}

	public String display(String prefix, int fieldFieldNr, String value) throws Exception {
		String fieldValue = value == null ? "" : new StringBuilder().append("[").append(value).append("]").toString();

		// String format = prefix + lengthFormatter.getDescription()
		// + validator.getDescription() + lengthFormatter.getMaxLength()
		// + formatter.getPackedLength(value == null ? 0 : value.length())
		// + fieldFieldNr + fieldValue
		// + System.getProperty("line.seperator");

		// StringBuilder sb = new StringBuilder();
		// sb.append(String.format("'%' ['%$-8'} '%$-4'} '%$6'} '%$d4'] '%'",
		// prefix, lengthFormatter.getDescription(),
		// validator.getDescription(), lengthFormatter.getMaxLength(),
		// formatter.getPackedLength(value == null ? 0 : value.length()),
		// fieldFieldNr, fieldValue, System.getProperty("line.seperator")));

		StringBuilder newSb = new StringBuilder();
		newSb.append(prefix).append(Utils.padRight(lengthFormatter.getDescription(), 8, ' '));
		newSb.append(Utils.padRight(validator.getDescription(), 4, ' '));
		newSb.append(Utils.padRight(lengthFormatter.getMaxLength(), 6, ' '));
		newSb.append(Utils.padRight(Integer.toString(formatter.getPackedLength(value == null ? 0 : value.length())), 4,
				' '));
		newSb.append(fieldFieldNr);
		newSb.append(fieldValue);
		newSb.append("\n");
		return newSb.toString();

		// String formatted = String.format(
		// "{0} [{1,-8} {2, -4} {3,6} {4:d4}] {5:d3} {6}", prefix,
		// lengthFormatter.getDescription(), validator.getDescription(),
		// lengthFormatter.getMaxLength(),
		// formatter.getPackedLength(value == null ? 0 : value.length()),
		// fieldFieldNr, fieldValue);

		// return sb.toString();
	}

	public Adjuster getAdjuster() {
		return adjuster;
	}

	public IFormatter getFormatter() {
		return formatter;
	}

	public ILengthFormatter getLengthFormatter() {
		return lengthFormatter;
	}

	public int getPackedLenghth(String value) throws Exception {
		return lengthFormatter.getLengthOfLengthIndicator() + formatter.getPackedLength(value.length());
	}

	public IFieldValidator getValidator() {
		return validator;
	}

	public byte[] pack(int fieldNumber, String value) throws Exception {
		if (!lengthFormatter.isValidLength(formatter.getPackedLength(value.length())))
			throw new FieldFormatException(fieldNumber, "The field length is not valid");

		if (!validator.isValid(value))
			throw new FieldFormatException(fieldNumber, "Invalid value for field");

		int lenOfLenInd = lengthFormatter.getLengthOfLengthIndicator();
		int lenOfField = formatter.getPackedLength(value.length());

		byte[] field = new byte[lenOfLenInd + lenOfField];
		lengthFormatter.pack(field, value.length(), 0);

		byte[] fieldData = formatter.getBytes(value);
		System.arraycopy(fieldData, 0, field, lenOfLenInd, lenOfField);
		return field;
	}

	public UnpackObject unpack(int fieldNumber, byte[] data, int offset) throws Exception {
		int lenOfLenInd = lengthFormatter.getLengthOfLengthIndicator();
		int lenOfField = lengthFormatter.getLengthOfField(data, offset);
		if (formatter instanceof BcdFormatter)
			lenOfField = formatter.getPackedLength(lenOfField);

		byte[] fieldData = new byte[lenOfField];
		System.arraycopy(data, offset + lenOfLenInd, fieldData, 0, lenOfField);
		int newOffset = offset + lenOfField + lenOfLenInd;

		String value = formatter.getString(fieldData);

		if (!validator.isValid(value))
			throw new FieldFormatException(fieldNumber, "Invalid field format");

		int length = value.length();

		if (formatter instanceof BinaryFormatter)
			length = formatter.getPackedLength(length);
		if (!lengthFormatter.isValidLength(length))
			throw new FieldFormatException(fieldNumber, "Field is too long");

		return new UnpackObject(value, newOffset);

	}

}
